# SPDX-License-Identifier: GPL-2.0-or-later

from typing import Optional
from gi.repository import Gio  # noqa

import configparser


def get_svg(model: str) -> Optional[bytes]:
    resource = Gio.resources_lookup_data(
        "/org/freedesktop/Piper/svgs/svg-lookup.ini", Gio.ResourceLookupFlags.NONE
    )

    data = resource.get_data()
    assert data is not None
    config = configparser.ConfigParser()
    config.read_string(data.decode("utf-8"), source="svg-lookup.ini")
    assert config.sections()

    filename = "fallback.svg"

    if model.startswith(("usb:", "bluetooth:")):
        bus, vid, pid, version = model.split(":")
        # Where the version is 0 (virtually all devices) we drop it. This
        # way the DeviceMatch lines are less confusing.
        usbid = ":".join([bus, vid, pid]) if int(version) == 0 else model

        for s in config.sections():
            matches = config[s]["DeviceMatch"].split(";")
            if usbid in matches:
                filename = config[s]["Svg"]
                break

    resource = Gio.resources_lookup_data(
        f"/org/freedesktop/Piper/svgs/{filename}", Gio.ResourceLookupFlags.NONE
    )

    return resource.get_data()



# SPDX-License-Identifier: GPL-2.0-or-later

from gettext import gettext as _
from typing import Callable, List, Optional

from .errorperspective import ErrorPerspective
from .mouseperspective import MousePerspective
from .welcomeperspective import WelcomePerspective
from .ratbagd import (
    Ratbagd,
    RatbagdDevice,
    RatbagdIncompatibleError,
    RatbagdUnavailableError,
)

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, GLib, Gtk, Gio  # noqa


@Gtk.Template(resource_path="/org/freedesktop/Piper/ui/Window.ui")
class Window(Gtk.ApplicationWindow):
    """A Gtk.ApplicationWindow subclass to implement the main application
    window. This window displays the different perspectives (error, mouse and
    welcome) that each present their own behavior."""

    __gtype_name__ = "Window"

    primary_menu: Gtk.Menu = Gtk.Template.Child()  # type: ignore
    stack_perspectives: Gtk.Stack = Gtk.Template.Child()  # type: ignore
    stack_titlebar: Gtk.Stack = Gtk.Template.Child()  # type: ignore

    def __init__(self, init_ratbagd_cb: Callable[[], Ratbagd], *args, **kwargs) -> None:
        """Instantiates a new Window.

        @param ratbag The ratbag instance to connect to, as ratbagd.Ratbag
        """
        Gtk.ApplicationWindow.__init__(self, *args, **kwargs)

        self.set_icon_name("org.freedesktop.Piper")

        self._add_perspective(ErrorPerspective(), None)
        try:
            ratbag = init_ratbagd_cb()
        except RatbagdUnavailableError:
            self._present_error_perspective(
                _("Cannot connect to ratbagd"),
                _(
                    "Please make sure ratbagd is running and your user is in the required group"
                ),
            )
            return
        except RatbagdIncompatibleError as e:
            self._present_error_perspective(
                _(
                    f"Incompatible ratbagd API version (required: {e.required_version}, provided: {e.ratbagd_version})"
                ),
                _("Please update both piper and libratbag to the latest versions"),
            )
            return

        for perspective in [MousePerspective(), WelcomePerspective()]:
            self._add_perspective(perspective, ratbag)

        welcome_perspective: WelcomePerspective = self._get_child("welcome_perspective")  # type: ignore
        welcome_perspective.connect("device-selected", self._on_device_selected)

        ratbag.connect("device-added", self._on_device_added)
        ratbag.connect("device-removed", self._on_device_removed)
        ratbag.connect("daemon-disappeared", self._on_daemon_disappeared)

        if len(ratbag.devices) == 0:
            self._present_error_perspective(
                _("Cannot find any devices"),
                _("Please make sure your device is supported and plugged in"),
            )
        elif len(ratbag.devices) == 1:
            self._present_mouse_perspective(ratbag.devices[0])
        else:
            self._present_welcome_perspective(ratbag.devices)

    def do_delete_event(self, event: Gdk.Event) -> bool:
        for perspective in self.stack_perspectives.get_children():
            if not perspective.can_shutdown:
                dialog = Gtk.MessageDialog(
                    self,
                    Gtk.DialogFlags.MODAL,
                    Gtk.MessageType.QUESTION,
                    Gtk.ButtonsType.YES_NO,
                    _("There are unapplied changes. Are you sure you want to quit?"),
                )
                response = dialog.run()
                dialog.destroy()

                if response in [Gtk.ResponseType.NO, Gtk.ResponseType.DELETE_EVENT]:
                    return Gdk.EVENT_STOP
        return Gdk.EVENT_PROPAGATE

    def _on_daemon_disappeared(self, ratbag: Ratbagd) -> None:
        self._present_error_perspective(
            _("Ooops. ratbagd has disappeared"), _("Please restart Piper")
        )

    def _on_device_added(self, ratbag: Ratbagd, device: RatbagdDevice) -> None:
        if len(ratbag.devices) == 1:
            # We went from 0 devices to 1 device; immediately configure it.
            self._present_mouse_perspective(device)
        elif self.stack_perspectives.get_visible_child_name() == "welcome_perspective":
            # We're in the welcome perspective; just add it to the list.
            welcome_perspective: WelcomePerspective = self._get_child(
                "welcome_perspective"
            )  # type: ignore
            welcome_perspective.add_device(device)
        else:
            # We're configuring another device; just notify the user.
            # TODO: show in-app notification?
            print("Device connected")

    def _on_device_removed(self, ratbag: Ratbagd, device: RatbagdDevice) -> None:
        mouse_perspective: MousePerspective = self._get_child("mouse_perspective")  # type: ignore

        if device is mouse_perspective.device:
            # The current device disconnected, which can only happen from the
            # mouse perspective as we'd otherwise be in the welcome screen with
            # more than one device remaining. Hence, we display the error
            # perspective.
            self._present_error_perspective(
                _("Your device disconnected!"),
                _("Please make sure your device is plugged in"),
            )
        elif self.stack_perspectives.get_visible_child_name() == "welcome_perspective":
            # We're in the welcome screen; just remove it from the list. If
            # there is nothing left, display the error perspective.
            welcome_perspective: WelcomePerspective = self._get_child("welcome_perspective")  # type: ignore
            welcome_perspective.remove_device(device)
            if len(ratbag.devices) == 0:
                self._present_error_perspective(
                    _("Cannot find any devices"),
                    _("Please make sure your device is supported and plugged in"),
                )
        else:
            # We're configuring another device; just notify the user.
            # TODO: show in-app notification?
            print("Device disconnected")

    def _add_perspective(self, perspective, ratbag: Optional[Ratbagd]) -> None:
        self.stack_perspectives.add_named(perspective, perspective.name)
        self.stack_titlebar.add_named(perspective.titlebar, perspective.name)
        if perspective.can_go_back:
            if ratbag is None:
                raise ValueError(
                    "`ratbag` must be provided if the perspective can go back"
                )
            self._perspective_add_back_button(perspective, ratbag)
        self._perspective_add_primary_menu(perspective)

    def _present_welcome_perspective(self, devices: List[RatbagdDevice]) -> None:
        # Present the welcome perspective for the user to select one of their
        # devices.
        welcome_perspective: WelcomePerspective = self._get_child("welcome_perspective")  # type: ignore
        welcome_perspective.set_devices(devices)

        self.stack_titlebar.set_visible_child_name(welcome_perspective.name)
        self.stack_perspectives.set_visible_child_name(welcome_perspective.name)

    def _present_mouse_perspective(self, device: RatbagdDevice) -> None:
        # Present the mouse configuration perspective for the given device.
        try:
            mouse_perspective: MousePerspective = self._get_child("mouse_perspective")  # type: ignore
            mouse_perspective.set_device(device)

            self.stack_titlebar.set_visible_child_name(mouse_perspective.name)
            self.stack_perspectives.set_visible_child_name(mouse_perspective.name)
        except ValueError as e:
            self._present_error_perspective(_("Cannot display device SVG"), str(e))
        except GLib.Error as e:
            # Happens with the GetSvgFd() call when running against older
            # python. This can be removed when we've had the newer call out
            # for a while. The full error is printed to stderr by
            # ratbagd.py.
            if e.code == Gio.DBusError.UNKNOWN_METHOD:
                self._present_error_perspective(
                    _("Newer version of ratbagd required"),
                    _("Please update to the latest available version"),
                )
            else:
                self._present_error_perspective(
                    _("Unknown exception occurred"), e.message
                )

    def _present_error_perspective(self, message: str, detail: str) -> None:
        # Present the error perspective informing the user of any errors.
        error_perspective: ErrorPerspective = self._get_child("error_perspective")  # type: ignore
        error_perspective.set_message(message)
        error_perspective.set_detail(detail)

        self.stack_titlebar.set_visible_child_name(error_perspective.name)
        self.stack_perspectives.set_visible_child_name(error_perspective.name)

    def _on_device_selected(self, perspective, device: RatbagdDevice) -> None:
        self._present_mouse_perspective(device)

    def _get_child(self, name: str) -> Gtk.Widget:
        child = self.stack_perspectives.get_child_by_name(name)
        if child is None:
            raise ValueError(f"Child `{name}` was not found")
        return child

    def _perspective_add_back_button(self, perspective, ratbag: Ratbagd) -> None:
        button_back = Gtk.Button.new_from_icon_name(
            "go-previous-symbolic", Gtk.IconSize.BUTTON
        )
        button_back.set_visible(len(ratbag.devices) > 1)
        button_back.connect(
            "clicked",
            lambda _, ratbag: self._present_welcome_perspective(ratbag.devices),
            ratbag,
        )
        ratbag.connect(
            "notify::devices",
            lambda ratbag, _: button_back.set_visible(len(ratbag.devices) > 1),
        )
        perspective.titlebar.add(button_back)
        # Place the button first in the titlebar.
        perspective.titlebar.child_set_property(button_back, "position", 0)

    def _perspective_add_primary_menu(self, perspective) -> None:
        hamburger = Gtk.Image.new_from_icon_name(
            "open-menu-symbolic", Gtk.IconSize.BUTTON
        )
        hamburger.set_visible(True)
        button_primary_menu = Gtk.MenuButton.new()
        button_primary_menu.add(hamburger)
        button_primary_menu.set_visible(True)
        button_primary_menu.set_menu_model(self.primary_menu)
        perspective.titlebar.pack_end(button_primary_menu)
        # Place the button last in the titlebar.
        perspective.titlebar.child_set_property(button_primary_menu, "position", 0)



# SPDX-License-Identifier: GPL-2.0-or-later

from typing import Optional

from .ratbagd import Ratbagd
from .window import Window

import gi

gi.require_version("Gio", "2.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Gio, GLib, Gtk  # noqa


class Application(Gtk.Application):
    """A Gtk.Application subclass to handle the application's initialization and
    integration with the GNOME stack. It implements the do_startup and
    do_activate methods and is responsible for the application's menus, icons,
    title and lifetime."""

    def __init__(self, ratbagd_api_version: int) -> None:
        """Instantiates a new Application."""
        Gtk.Application.__init__(
            self,
            application_id="org.freedesktop.Piper",
            flags=Gio.ApplicationFlags.FLAGS_NONE,
        )
        GLib.set_application_name("Piper")
        self._required_ratbagd_version = ratbagd_api_version

    def do_startup(self) -> None:
        """This function is called when the application is first started. All
        initialization should be done here, to prevent doing duplicate work in
        case another window is opened."""
        Gtk.Application.do_startup(self)
        self._build_app_menu()
        self._ratbagd: Optional[Ratbagd] = None

    def init_ratbagd(self) -> Ratbagd:
        if self._ratbagd is None:
            self._ratbag = Ratbagd(self._required_ratbagd_version)
        return self._ratbag

    def do_activate(self) -> None:
        """This function is called when the user requests a new window to be
        opened."""
        window = Window(self.init_ratbagd, application=self)
        window.present()

    def _build_app_menu(self) -> None:
        # Set up the app menu
        actions = [("about", self._about), ("quit", self._quit)]
        for name, callback in actions:
            action = Gio.SimpleAction.new(name, None)
            action.connect("activate", callback)
            self.add_action(action)

    def _about(self, action: Gio.SimpleAction, param: None) -> None:
        # Set up the about dialog.
        builder = Gtk.Builder().new_from_resource(
            "/org/freedesktop/Piper/AboutDialog.ui"
        )
        about: Optional[Gtk.AboutDialog] = builder.get_object("about_dialog")  # type: ignore
        assert about is not None
        about.set_transient_for(self.get_active_window())
        about.connect("response", lambda about, param: about.destroy())
        about.show()

    def _quit(self, action: Gio.SimpleAction, param: None) -> None:
        # Quit the application.
        windows = self.get_windows()
        for window in windows:
            window.destroy()



# SPDX-License-Identifier: GPL-2.0-or-later

from gettext import gettext as _
from typing import Optional

from .buttonspage import ButtonsPage
from .profilerow import ProfileRow
from .ratbagd import RatbagdDevice, RatbagdProfile
from .resolutionspage import ResolutionsPage
from .advancedpage import AdvancedPage
from .ledspage import LedsPage
from .util.gobject import connect_signal_with_weak_ref

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import GLib, GObject, Gtk  # noqa


@Gtk.Template(resource_path="/org/freedesktop/Piper/ui/MousePerspective.ui")
class MousePerspective(Gtk.Overlay):
    """The perspective to configure a mouse."""

    __gtype_name__ = "MousePerspective"

    _titlebar: Gtk.HeaderBar = Gtk.Template.Child()  # type: ignore
    add_profile_button: Gtk.Button = Gtk.Template.Child()  # type: ignore
    button_commit: Gtk.Button = Gtk.Template.Child()  # type: ignore
    button_profile: Gtk.Button = Gtk.Template.Child()  # type: ignore
    label_profile: Gtk.Label = Gtk.Template.Child()  # type: ignore
    listbox_profiles: Gtk.ListBox = Gtk.Template.Child()  # type: ignore
    notification_error: Gtk.Revealer = Gtk.Template.Child()  # type: ignore
    stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore

    def __init__(self, *args, **kwargs) -> None:
        """Instantiates a new MousePerspective."""
        Gtk.Overlay.__init__(self, *args, **kwargs)
        self._device: Optional[RatbagdDevice] = None
        self._profile: Optional[RatbagdProfile] = None
        self._notification_error_timeout_id = 0

    @GObject.Property
    def name(self) -> str:
        """The name of this perspective."""
        return "mouse_perspective"

    @GObject.Property
    def titlebar(self) -> Gtk.Widget:
        """The titlebar to this perspective."""
        return self._titlebar

    @GObject.Property
    def can_go_back(self) -> bool:
        """Whether this perspective wants a back button to be displayed in case
        there is more than one connected device."""
        return True

    @GObject.Property
    def can_shutdown(self) -> bool:
        if self._device is None:
            return True

        """Whether this perspective can safely shutdown."""
        return all(not profile.dirty for profile in self._device.profiles)

    @GObject.Property
    def device(self) -> RatbagdDevice:
        assert self._device is not None
        return self._device

    def set_device(self, device: RatbagdDevice) -> None:
        self._device = device
        connect_signal_with_weak_ref(
            self, device, "resync", lambda _: self._show_notification_error()
        )
        connect_signal_with_weak_ref(
            self,
            self._device,
            "active-profile-changed",
            self._on_active_profile_changed,
        )

        active_profile = device.active_profile
        assert active_profile is not None
        self._set_profile(active_profile)

        self.button_profile.set_visible(len(device.profiles) > 1)

        self.listbox_profiles.foreach(Gtk.Widget.destroy)
        for profile in device.profiles:
            connect_signal_with_weak_ref(
                self, profile, "notify::disabled", self._on_profile_notify_disabled
            )
            connect_signal_with_weak_ref(
                self, profile, "notify::dirty", self._on_profile_notify_dirty
            )
            row = ProfileRow(profile)
            self.listbox_profiles.insert(row, profile.index)

        self._on_profile_notify_disabled(active_profile, None)

        self._select_profile_row(active_profile)

    def _select_profile_row(self, profile: RatbagdProfile) -> None:
        for row in self.listbox_profiles.get_children():
            if row.profile is profile:
                self.listbox_profiles.select_row(row)
                self.label_profile.set_label(row.name)
                break

    def _set_profile(self, profile: RatbagdProfile) -> None:
        assert self._device is not None

        self._select_profile_row(profile)

        self._profile = profile

        self.stack.foreach(Gtk.Widget.destroy)
        if profile.resolutions:
            self.stack.add_titled(
                ResolutionsPage(self._device, profile), "resolutions", _("Resolutions")
            )
        if profile.buttons:
            self.stack.add_titled(
                ButtonsPage(self._device, profile), "buttons", _("Buttons")
            )
        if profile.leds:
            self.stack.add_titled(LedsPage(self._device, profile), "leds", _("LEDs"))
        # TODO: get rid of this duplicated logic.
        are_report_rates_supported = (
            profile.report_rate != 0 and len(profile.report_rates) != 0
        )
        if (
            profile.angle_snapping != -1
            or profile.debounces
            or are_report_rates_supported
        ):
            self.stack.add_titled(
                AdvancedPage(self._device, profile), "advanced", _("Advanced")
            )

        self._on_profile_notify_dirty(profile, None)

    def _hide_notification_error(self) -> None:
        if self._notification_error_timeout_id != 0:
            GLib.Source.remove(self._notification_error_timeout_id)
            self._notification_error_timeout_id = 0
        self.notification_error.set_reveal_child(False)

    def _show_notification_error(self) -> None:
        self.notification_error.set_reveal_child(True)
        self._notification_error_timeout_id = GLib.timeout_add_seconds(
            5, self._on_notification_error_timeout
        )

    def _on_active_profile_changed(
        self, _device: RatbagdDevice, profile: RatbagdProfile
    ) -> None:
        # TODO: preserve the active tab.
        self._set_profile(profile)

    def _on_notification_error_timeout(self) -> bool:
        self._hide_notification_error()
        return False

    @Gtk.Template.Callback("_on_save_button_clicked")
    def _on_save_button_clicked(self, _button: Gtk.Button) -> None:
        assert self._device is not None
        self._device.commit()

    @Gtk.Template.Callback("_on_notification_error_close_clicked")
    def _on_notification_error_close_clicked(self, button: Gtk.Button) -> None:
        self._hide_notification_error()

    @Gtk.Template.Callback("_on_profile_row_activated")
    def _on_profile_row_activated(self, listbox: Gtk.ListBox, row: ProfileRow) -> None:
        row.set_active()

    @Gtk.Template.Callback("_on_add_profile_button_clicked")
    def _on_add_profile_button_clicked(self, button: Gtk.Button) -> None:
        assert self._device is not None
        # Enable the first disabled profile we find.
        for profile in self._device.profiles:
            if not profile.disabled:
                continue
            profile.disabled = False
            break

    def _on_profile_notify_disabled(
        self, profile: RatbagdProfile, pspec: Optional[GObject.ParamSpec]
    ) -> None:
        assert self._device is not None

        self.add_profile_button.set_sensitive(
            any(p.disabled for p in self._device.profiles)
        )

    def _on_profile_notify_dirty(
        self, profile: RatbagdProfile, pspec: Optional[GObject.ParamSpec]
    ) -> None:
        assert self._device is not None

        device_dirty = any(p.dirty for p in self._device.profiles)

        style_context = self.button_commit.get_style_context()
        if device_dirty:
            style_context.add_class("suggested-action")
            self.button_commit.set_sensitive(True)
        else:
            # There is no way to make a single profile non-dirty, so this works
            # for now. Ideally, this should however check if there are any other
            # profiles on the device that are dirty.
            style_context.remove_class("suggested-action")
            self.button_commit.set_sensitive(False)



# SPDX-License-Identifier: GPL-2.0-or-later

import sys
from typing import Optional
import gi

from .mousemap import MouseMap
from .ratbagd import RatbagdDevice, RatbagdProfile
from .util.gobject import connect_signal_with_weak_ref

gi.require_version("Gtk", "3.0")
from gi.repository import GObject, Gtk  # noqa: E402


@Gtk.Template(resource_path="/org/freedesktop/Piper/ui/AdvancedPage.ui")
class AdvancedPage(Gtk.Box):
    """
    Advanced settings stack.
    """

    __gtype_name__ = "AdvancedPage"

    angle_snapping: Gtk.Switch = Gtk.Template.Child()  # type: ignore
    debounce: Gtk.ComboBox = Gtk.Template.Child()  # type: ignore
    rate_1000: Gtk.RadioButton = Gtk.Template.Child()  # type: ignore
    rate_125: Gtk.RadioButton = Gtk.Template.Child()  # type: ignore
    rate_250: Gtk.RadioButton = Gtk.Template.Child()  # type: ignore
    rate_500: Gtk.RadioButton = Gtk.Template.Child()  # type: ignore
    rate_button_box: Gtk.ButtonBox = Gtk.Template.Child()  # type: ignore

    def __init__(
        self, device: RatbagdDevice, profile: RatbagdProfile, *args, **kwargs
    ) -> None:
        """Instantiates a new AdvancedPage."""
        Gtk.Box.__init__(self, *args, **kwargs)

        self._profile = profile

        cell = Gtk.CellRendererText()
        self.debounce.pack_start(cell, True)
        self.debounce.add_attribute(cell, "text", 0)

        model = Gtk.ListStore(str)
        for ms in profile.debounces:
            model.append([str(ms)])
        self.debounce.set_model(model)

        self._handler_debounce = self.debounce.connect(
            "changed", self._on_debounce_combo_changed
        )

        self._profile_debounce_time_changed_handler = connect_signal_with_weak_ref(
            self,
            self._profile,
            "notify::debounce",
            self._on_profile_debounce_time_changed,
        )
        self._update_widget_debounce_time()

        are_report_rates_supported = (
            profile.report_rate != 0 and len(profile.report_rates) != 0
        )
        self.rate_button_box.set_sensitive(are_report_rates_supported)
        self.rate_125.set_sensitive(125 in profile.report_rates)
        self.rate_250.set_sensitive(250 in profile.report_rates)
        self.rate_500.set_sensitive(500 in profile.report_rates)
        self.rate_1000.set_sensitive(1000 in profile.report_rates)

        self._handler_125 = self.rate_125.connect(
            "toggled", self._on_report_rate_toggled, 125
        )
        self._handler_250 = self.rate_250.connect(
            "toggled", self._on_report_rate_toggled, 250
        )
        self._handler_500 = self.rate_500.connect(
            "toggled", self._on_report_rate_toggled, 500
        )
        self._handler_1000 = self.rate_1000.connect(
            "toggled", self._on_report_rate_toggled, 1000
        )

        self._profile_report_rate_changed_handler = connect_signal_with_weak_ref(
            self,
            self._profile,
            "notify::report-rate",
            self._on_profile_report_rate_changed,
        )
        self._update_widget_report_rate()

        profile = self._profile

        self._mousemap = MouseMap("#Buttons", device, spacing=20, border_width=20)
        self.pack_start(self._mousemap, True, True, 0)

        self._angle_snapping_switch_handler = self.angle_snapping.connect(
            "state-set", self._on_angle_snapping_switch_state_set
        )

        self.angle_snapping.set_sensitive(profile.angle_snapping != -1)

        self._profile_angle_snapping_changed_handler = connect_signal_with_weak_ref(
            self,
            self._profile,
            "notify::angle-snapping",
            self._on_profile_angle_snapping_changed,
        )
        self._update_widget_angle_snapping()

        self.show_all()

    def _on_profile_debounce_time_changed(
        self, profile: RatbagdProfile, pspec: Optional[GObject.ParamSpec]
    ) -> None:
        self._update_widget_debounce_time()

    def _update_widget_debounce_time(self) -> None:
        profile = self._profile

        with self.debounce.handler_block(self._handler_debounce):
            if profile.debounce in profile.debounces:
                idx = profile.debounces.index(profile.debounce)
                self.debounce.set_active(idx)
            else:
                self.debounce.set_active(0)

    def _on_debounce_combo_changed(self, combo: Gtk.ComboBox) -> None:
        idx = combo.get_active()
        profile = self._profile
        with profile.handler_block(self._profile_debounce_time_changed_handler):
            profile.debounce = profile.debounces[idx]

    def _on_angle_snapping_switch_state_set(
        self, button: Gtk.Switch, state: bool
    ) -> None:
        profile = self._profile
        with profile.handler_block(self._profile_angle_snapping_changed_handler):
            profile.angle_snapping = 1 if state else 0

    def _on_profile_angle_snapping_changed(
        self, profile: RatbagdProfile, pspec: Optional[GObject.ParamSpec]
    ) -> None:
        self._update_widget_angle_snapping()

    def _update_widget_angle_snapping(self) -> None:
        with self.angle_snapping.handler_block(self._angle_snapping_switch_handler):
            self.angle_snapping.set_active(self._profile.angle_snapping == 1)

    def _on_report_rate_toggled(self, button: Gtk.RadioButton, rate: int) -> None:
        if not button.get_active():
            return
        profile = self._profile
        with profile.handler_block(self._profile_report_rate_changed_handler):
            profile.report_rate = rate

    def _on_profile_report_rate_changed(
        self, profile: RatbagdProfile, pspec: Optional[GObject.ParamSpec]
    ) -> None:
        self._update_widget_report_rate()

    def _update_widget_report_rate(self) -> None:
        with self.rate_125.handler_block(self._handler_125):
            if self._profile.report_rate == 125:
                self.rate_125.set_active(True)
                return

        with self.rate_250.handler_block(self._handler_250):
            if self._profile.report_rate == 250:
                self.rate_250.set_active(True)
                return

        with self.rate_500.handler_block(self._handler_500):
            if self._profile.report_rate == 500:
                self.rate_500.set_active(True)
                return

        with self.rate_1000.handler_block(self._handler_1000):
            if self._profile.report_rate == 1000:
                self.rate_1000.set_active(True)
                return

        # TODO: think how we should handle this.
        print(
            f"Profile was set to a weird report rate: {self._profile.report_rate}",
            file=sys.stderr,
        )



# SPDX-License-Identifier: GPL-2.0-or-later

from typing import Optional

import gi

from .ratbagd import RatbagdResolution
from .util.gobject import connect_signal_with_weak_ref

gi.require_version("Gtk", "3.0")
from gi.repository import GObject, Gdk, Gtk  # noqa


class _DPIEntry(Gtk.Entry, Gtk.Editable):
    """
    A subclass of Gtk.Entry with an overridden method for input validation.

    This class addresses a specific bug in the handling of parameters in pygobject when
    using signals. For Gtk.Entry, the 'position' parameter is incorrect when handling
    the 'insert-text' signal, causing an assertion warning to be logged. By using this
    subclass, we can override the 'insert-text' logic to correctly handle the position
    parameter and avoid the warning.

    Inheriting from Gtk.Editable in addition to Gtk.Entry ensures that the overridden
    method applies only to instances of this class, rather than affecting all
    Gtk.Entry objects in the application.
    """

    __gtype_name__ = "DPIEntry"

    def do_insert_text(self, new_text: str, new_text_length: int, position: int) -> int:
        """Overrides the default Gtk.Editable insert-text handler to validate numerical
        input."""
        if not new_text.isdigit():
            self.stop_emission("insert-text")
            return position

        self.get_buffer().insert_text(position, new_text, new_text_length)
        return position + new_text_length


@Gtk.Template(resource_path="/org/freedesktop/Piper/ui/ResolutionRow.ui")
class ResolutionRow(Gtk.ListBoxRow):
    """A Gtk.ListBoxRow subclass containing the widgets to configure a
    resolution."""

    __gtype_name__ = "ResolutionRow"

    active_button: Gtk.Button = Gtk.Template.Child()  # type: ignore
    active_label: Gtk.Label = Gtk.Template.Child()  # type: ignore
    disable_button: Gtk.Button = Gtk.Template.Child()  # type: ignore
    dpi_label: Gtk.Label = Gtk.Template.Child()  # type: ignore
    dpi_entry: _DPIEntry = Gtk.Template.Child()  # type: ignore
    revealer: Gtk.Revealer = Gtk.Template.Child()  # type: ignore
    scale: Gtk.Scale = Gtk.Template.Child()  # type: ignore

    CAP_SEPARATE_XY_RESOLUTION = False
    CAP_DISABLE = False

    def __init__(
        self, resolution: RatbagdResolution, resolutions_page, *args, **kwargs
    ) -> None:
        Gtk.ListBoxRow.__init__(self, *args, **kwargs)

        self.resolutions_page = resolutions_page
        self._resolution = resolution
        self.resolutions = resolution.resolutions
        self._scale_handler = self.scale.connect(
            "value-changed", self._on_scale_value_changed
        )
        self._disabled_button_handler = self.disable_button.connect(
            "toggled", self._on_disable_button_toggled
        )

        self._active_handler = connect_signal_with_weak_ref(
            self, resolution, "notify::is-active", self._on_status_changed
        )
        self._disabled_handler = connect_signal_with_weak_ref(
            self, resolution, "notify::is-disabled", self._on_status_changed
        )
        connect_signal_with_weak_ref(
            self, resolution, "notify::resolution", self._on_profile_resolution_changed
        )

        # Get resolution capabilities and update internal values.
        if RatbagdResolution.CAP_SEPARATE_XY_RESOLUTION in resolution.capabilities:
            self.CAP_SEPARATE_XY_RESOLUTION = True
        if RatbagdResolution.CAP_DISABLE in resolution.capabilities:
            self.CAP_DISABLE = True

        # Set initial values for the UI.
        res = resolution.resolution[0]
        minres = resolution.resolutions[0]
        maxres = resolution.resolutions[-1]
        res_num_chars = len(str(maxres))

        # Set the max width and length for the DPI text boxes, based on the max res value.
        self.dpi_entry.set_width_chars(res_num_chars)
        self.dpi_entry.set_max_length(res_num_chars)

        with self.scale.handler_block(self._scale_handler):
            self.scale.props.adjustment.configure(res, minres, maxres, 50, 50, 0)
            self.scale.set_value(res)
        if resolution.is_disabled:
            with self.disable_button.handler_block(self._disabled_button_handler):
                self.disable_button.set_active(True)
        self._on_status_changed(resolution, _pspec=None)

    @Gtk.Template.Callback("_on_change_value")
    def _on_change_value(
        self, scale: Gtk.Scale, scroll: Gtk.ScrollType, value: float
    ) -> bool:
        # Cursor-controlled slider may get out of the GtkAdjustment's range.
        value = min(max(self.resolutions[0], value), self.resolutions[-1])

        # Find the nearest permitted value to our Gtk.Scale value.
        lo = max(r for r in self.resolutions if r <= value)
        hi = min(r for r in self.resolutions if r >= value)

        value = lo if value - lo < hi - value else hi

        scale.set_value(value)
        self.dpi_entry.set_text(str(value))

        # libratbag provides a fake-exponential range with the deltas
        # increasing as the resolution goes up. Make sure we set our
        # steps to the next available value.
        idx = self.resolutions.index(value)
        if idx < len(self.resolutions) - 1:
            delta = self.resolutions[idx + 1] - self.resolutions[idx]
            scale.props.adjustment.set_step_increment(delta)
            scale.props.adjustment.set_page_increment(delta)

        return True

    def _on_disable_button_toggled(self, togglebutton: Gtk.Button) -> None:
        # The disable button has been toggled, update RatbagdResolution.
        self._resolution.set_disabled(togglebutton.get_active())

        # Update UI
        self._on_status_changed(self._resolution, _pspec=None)

    @Gtk.Template.Callback("_on_active_button_clicked")
    def _on_active_button_clicked(self, _togglebutton: Gtk.Button) -> None:
        # The set active button has been clicked, update RatbagdResolution.
        self._resolution.set_active()

    @Gtk.Template.Callback("_on_scroll_event")
    def _on_scroll_event(self, widget: Gtk.Widget, _event: Gdk.EventScroll) -> bool:
        # Prevent a scroll in the list to get caught by the scale.
        GObject.signal_stop_emission_by_name(widget, "scroll-event")
        return False

    def _on_scale_value_changed(self, scale: Gtk.Scale) -> None:
        # The scale has been moved, update RatbagdResolution's resolution.
        res = int(scale.get_value())
        self._on_dpi_values_changed(res=res)

    def _on_status_changed(
        self, resolution: RatbagdResolution, _pspec: Optional[GObject.ParamSpec]
    ) -> None:
        # The resolution's status changed, update UI.
        self._on_dpi_values_changed()
        if resolution.is_active:
            self.active_label.set_visible(True)
            self.active_button.set_sensitive(False)
            self.disable_button.set_sensitive(False)
        else:
            self.active_label.set_visible(False)
            self.active_button.set_sensitive(True)
            if self.CAP_DISABLE:
                with self.disable_button.handler_block(self._disabled_button_handler):
                    self.disable_button.set_sensitive(not resolution.is_default)
                    if resolution.is_disabled:
                        self.disable_button.set_active(True)
                        self.active_button.set_sensitive(False)
                        self.dpi_label.set_sensitive(False)
                        self.dpi_entry.set_sensitive(False)
                        self.scale.set_sensitive(False)
                    else:
                        self.disable_button.set_active(False)
                        self.dpi_label.set_sensitive(True)
                        self.dpi_entry.set_sensitive(True)
                        self.scale.set_sensitive(True)

    def toggle_revealer(self) -> None:
        # Toggles the revealer to show or hide the configuration widgets.
        reveal = not self.revealer.get_reveal_child()
        self.revealer.set_reveal_child(reveal)
        if reveal:
            self.dpi_entry.grab_focus()

    def _on_dpi_values_changed(self, res: Optional[int] = None) -> None:
        # Freeze the notify::resolution signal from firing and
        # update DPI text box and resolution values.
        if res is None:
            res = self._resolution.resolution[0]
        new_res = (res, res) if self.CAP_SEPARATE_XY_RESOLUTION else (res,)
        self.dpi_entry.set_text(str(res))

        # Only update new resolution if changed
        if new_res != self._resolution.resolution:
            self._resolution.resolution = new_res

    @Gtk.Template.Callback("_on_dpi_entry_activate")
    def _on_dpi_entry_activate(self, entry: Gtk.Entry) -> None:
        # The DPI entry has been changed, update the scale and RatbagdResolution's resolution.
        try:
            res = int(entry.get_text())
            # Get the resolution closest to what has been entered
            closest_res = min(self.resolutions, key=lambda x: abs(x - res))
            entry.set_text(str(closest_res))
            self._on_dpi_values_changed(res=closest_res)

            with self.scale.handler_block(self._scale_handler):
                self.scale.set_value(res)
        except ValueError:
            # If the input is not a valid integer, reset to the current value.
            entry.set_text(str(self._resolution.resolution[0]))

    @Gtk.Template.Callback("_on_dpi_entry_focus_in")
    def _on_dpi_entry_focus_in(self, _entry: Gtk.Entry, _event_focus: Gdk.EventFocus):
        if not self.revealer.get_reveal_child():
            self.resolutions_page.on_row_activated(None, self)

    @Gtk.Template.Callback("_on_dpi_entry_focus_out")
    def _on_dpi_entry_focus_out(self, entry: Gtk.Entry, _event_focus: Gdk.EventFocus):
        # Apply the DPI value on focus out
        self._on_dpi_entry_activate(entry)

    def _on_profile_resolution_changed(
        self, resolution: RatbagdResolution, _pspec: GObject.ParamSpec
    ) -> None:
        with self.scale.handler_block(self._scale_handler):
            res = resolution.resolution[0]
            self.scale.set_value(res)
            self.dpi_entry.set_text(str(res))



from typing import Callable, Union

from gi.repository import GObject


def connect_signal_with_weak_ref(
    ref_obj: Union[GObject.Object, GObject.GObject],
    obj: Union[GObject.Object, GObject.GObject],
    signal: str,
    func: Callable,
    *args,
) -> int:
    """
    Connect handler to an object `obj` tied to the life time of `ref_obj`.

    Use this to work around https://gitlab.gnome.org/GNOME/pygobject/-/issues/557.
    """

    handler = obj.connect(signal, func, *args)
    ref_obj.weak_ref(lambda: obj.disconnect(handler))
    return handler
